/*/../bin/true; p="/tmp/$(basename "$0").elf"; v=1; gcc -Wfatal-errors -std=c99 -O42 "$0" -o "$p" && chmod 755 "$p" && "$p" && v=0 || v=$?; rm "$p"; exit $v;*/


/*
 * yprof - y as in why
 * v1.1, 2020-02-28, ed <irc.rizon.net>, MIT-licensed
 * https://ocv.me/dev/?yprof.c
 * 
 * usage (see test_yprof_2):
 *   initialize with `yprof_vec *wv = yprof_new(24, true);`
 *   (where 24 is the initial vector size, number of measurements)
 *   then do measurements with YP and finally print the report with
 *   `yprof_report(wv);`
 */


// config: number of lines to show in report
#define YPROF_TRUNC 10


// convenience: take a sample
#define YP yprof_put(wv, __LINE__)


// just prints the current monotonic timestamp, no mallocs
#define YPM yprof_mono(__LINE__)
//#define YPM ;


// ok here we go
//


#define _POSIX_C_SOURCE 199309L

#include <stdbool.h>
#include <inttypes.h>
#include <stdlib.h>
#include <stdint.h>
#include <stdio.h>
#include <time.h>
#include <sys/times.h>


typedef struct yprof_sample {
	time_t sec;
	long nsec;
	clock_t cpu_usr;
	clock_t cpu_sys;
	int lineno;
} yprof_sample;


typedef struct yprof_vec {
	size_t size;
	size_t ubound;
	bool use_cpu;
	yprof_sample** data;
} yprof_vec;


void yprof_mono(const int lineno) {
	struct timespec tp;
	clock_gettime(CLOCK_MONOTONIC, &tp);
	fprintf(stdout, "%jd.%09ld L%d\n", tp.tv_sec, tp.tv_nsec, lineno);
}


int yprof_cmp(const void* a, const void* b) {
	const struct yprof_sample* sa = *(const struct yprof_sample**)a;
	const struct yprof_sample* sb = *(const struct yprof_sample**)b;
	
	const int64_t ret =
		sa->sec != sb->sec ?
			sa->sec - sb->sec :
			sa->nsec - sb->nsec;
	
	if (ret < 0) return 1;
	if (ret > 0) return -1;
	return 0;
}


struct yprof_vec* yprof_new(const size_t sz, const bool use_cpu) {
	struct yprof_vec* wv = malloc(sizeof(struct yprof_vec));
	wv->size = 0;
	wv->ubound = sz;
	wv->use_cpu = use_cpu;
	wv->data = malloc(wv->ubound * sizeof(struct yprof_sample));
	if (wv->data)
		return wv;
	
	return 0;
}


void yprof_del(struct yprof_vec* wv) {
	for (size_t a = 0; a < wv->size; ++a)
		free(wv->data[a]);
	
	free(wv->data);
	free(wv);
}


int yprof_put(struct yprof_vec* wv, const int lineno) {
	struct timespec tp;
	clock_gettime(CLOCK_MONOTONIC, &tp);
	
	if (wv->size >= wv->ubound) {
		wv->ubound *= 2;
		wv->data = realloc(wv->data, wv->ubound * sizeof(struct yprof_sample));
		if (!wv->data)
			return -1;
	}
	
	struct yprof_sample* s = malloc(sizeof(struct yprof_sample));
	
	s->sec = tp.tv_sec;
	s->nsec = tp.tv_nsec;
	s->lineno = lineno;

	if (wv->use_cpu) {
		struct tms cpu;
		times(&cpu);
		s->cpu_usr = cpu.tms_utime + cpu.tms_cutime;
		s->cpu_sys = cpu.tms_stime + cpu.tms_cstime;
	}
	
	wv->data[wv->size++] = s;
	return 0;
}


void yprof_report(struct yprof_vec* wv) {
	const bool use_cpu = wv->use_cpu;
	for (size_t a = wv->size - 1; a > 0; --a) {
		const struct yprof_sample* s0 = wv->data[a-1];
		struct yprof_sample* s = wv->data[a];
		
		if (s->nsec >= s0->nsec) {
			s->sec -= s0->sec;
			s->nsec -= s0->nsec;
		}
		else {
			s->nsec = (1000000000LL + s->nsec) - s0->nsec;
			s->sec = (s->sec - s0->sec) - 1;
		}
		
		if (use_cpu) {
			s->cpu_usr -= s0->cpu_usr;
			s->cpu_sys -= s0->cpu_sys;
		}
	}
	
	qsort(wv->data + 1, wv->size - 1, sizeof(struct yprof_sample*), yprof_cmp);
	
	const struct yprof_sample* sh = wv->data[0];
	if (use_cpu)
		fprintf(stdout, "%ld'%09ld L%d %ld %ld\n",
			sh->sec, sh->nsec, sh->lineno, sh->cpu_usr, sh->cpu_sys);
	else
		fprintf(stdout, "%ld'%09ld L%d\n",
			sh->sec, sh->nsec, sh->lineno);

	const size_t end = wv->size > YPROF_TRUNC ? YPROF_TRUNC : wv->size;
	for (size_t a = 1; a < end; ++a) {
		const yprof_sample s = *wv->data[a];
		if (use_cpu)
			fprintf(stdout, "%12lld L%d %ld %ld\n",
				s.sec * 1000000000LL + s.nsec, s.lineno, s.cpu_usr, s.cpu_sys);
		else
			fprintf(stdout, "%12lld L%d\n", s.sec * 1000000000LL + s.nsec, s.lineno);
			// stderr = ~260% cost
	}
	yprof_del(wv);
}


void test_yprof_1(const bool use_cpu) {
	yprof_vec *wv = yprof_new(24, use_cpu);
	YP;
	YP;
	YP; int a = 3;
	YP;
	YP;
	YP; a += 2;
	YP;
	YP;
	YP;
	if (a > 3) {
		YP;
		YP;
		YP;
	}
	YP;
	YP;
	YP;
	if (a < 3) {
		YP;
		YP;
		YP;
	}
	YP;
	YP;
	YP; a -= 2;
	YP;
	YP;
	YP; yprof_report(wv);
	//yprof_del(wv);
}


void test_yprof_2(const bool use_cpu) {
	yprof_vec *wv = yprof_new(24, use_cpu);
	YP; // initial sample; used for header
	YP; // next YP will be ~2x faster thanks to cpu cache
	YP; nanosleep((const struct timespec[]){{0, 1000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 2000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 3000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 4000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 5000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 6000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 7000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 8000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 9000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 10000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 11000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 12000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 13000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 14000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 15000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 16000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 17000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 18000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 19000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 20000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 21000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 22000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 23000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 24000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 25000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 26000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 27000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 28000000L}}, NULL);
	YP; nanosleep((const struct timespec[]){{0, 29000000L}}, NULL);
	YP; yprof_report(wv);
	//  3295547 L263 0 0
	//  2933208 L262 0 0
	//  1060611 L261 0 0
	//    17229 L259 0 0
	//     3880 L260 0 0
}


void test_yprof_1_loop(const bool use_cpu) {
	for (size_t a = 0; a < 500000; ++a)
		test_yprof_1(use_cpu);
	
	// YPROF_TRUNC 5
	//
	// alpine 3.11 (linux 5.4.12) musl gcc 9.2.0 on ryzen 3700x (LTSC 1809 vbox vm)
	// wall 1.885   user 1.834   sys 0.046   500000 without cpu
	// wall 3.306   user 0.985   sys 2.251    50000 with proc
	// wall 3.652   user 2.384   sys 1.267   500000 tms true
	// wall 1.696   user 1.692   sys 0.000   500000 tms false
	//
	// arch (linux 4.19.102) glibc gcc 9.2.0 on i7-4800MQ
	// wall 0.988   user 0.988   sys 0.000   500000 without cpu
	// wall 6.168   user 2.348   sys 3.715    50000 with proc
	// wall 6.906   user 4.002   sys 2.900   500000 tms true
	// wall 0.928   user 0.928   sys 0.000   500000 tms false
}


void test_getcpu_1() {
	clock_t x1, x2;
	for (int a = 0; a < 500000; a++) {
		struct tms cpu;
		times(&cpu);
		x1 = cpu.tms_utime + cpu.tms_cutime;
		x2 = cpu.tms_stime + cpu.tms_cstime;
	}
	// wall 0.268   user 0.157   sys 0.111   arch
}


int main(int argc, char** argv) {
	//test_yprof_1_loop(true);
	//test_yprof_1();
	test_yprof_2(true);
	//test_getcpu_1();
	return 0;
}


/* test stuff

TIMEFORMAT="wall %3R   user %3U   sys %3S"; gcc -Wfatal-errors -std=c99 -O42 yprof.c && for ((n=0; n<10; n++)); do time ./a.out > /dev/null; done

while true; do t=$(stat -c%Y yprof.c); sleep 0.2; [ $t = "$ot" ] && continue; ot="$t"; printf '\n\n\n\n\n'; gcc -Wfatal-errors -std=c99 -O0 yprof.c && valgrind --leak-check=full --track-origins=yes --partial-loads-ok=no --expensive-definedness-checks=yes --exit-on-first-error=yes --error-exitcode=1 ./a.out; done

c=a; gcc -Wfatal-errors -std=c99 -O42 yprof.c && for n in {01..20}; do echo -n "$n "; ./a.out > $c$n; done; echo
pr -mtW$(tput cols) a01 a02 a03 b01 b02 b03

*/
